from apalib.AminoAcid import AminoAcid
from apalib.Atom import Atom
from apalib.DNA import DNA
from apalib.RNA import RNA
from apalib.HETATM import HETATM

class Container:
    def __init__(self):
        self.current_fetch = None
        self.PeptideChains = None
        self.DNAChains = None
        self.RNAChains = None
        self.HETATMChains = None

    def ClearAll(self):
        self.__init__()

    def SetFetch(self, fetch):
        self.current_fetch = fetch

    def GetFetch(self):
        return self.current_fetch

    def ClearFetch(self):
        self.current_fetch = None

    def SetProteinChains(self, pchain):
        self.PeptideChains = pchain

    def GetPeptideChains(self):
        return self.PeptideChains

    def ClearPeptideChains(self):
        self.PeptideChains = None

    def SetDNAChains(self,dchain):
        self.DNAChains = dchain

    def GetDNAChains(self):
        return self.DNAChains

    def ClearDNAChains(self):
        self.DNAChains = None

    def GetRNAChains(self):
        return self.RNAChains

    def SetRNAChains(self, rchain):
        self.RNAChains = rchain

    def ClearRNAChains(self):
        self.RNAChains = None

    def GetHETATMChains(self):
        return self.HETATMChains

    def SetHETATMChains(self, hchain):
        self.HETATMChains = hchain

    def ClearHEETATMChains(self):
        self.HETATMChains = None

    #Return all residues from all chains as a single list
    def AsList(self, ordered=True):
        fullLst = []
        retLst = []
        lst = [self.PeptideChains, self.DNAChains, self.RNAChains, self.HETATMChains]
        for val in lst:
            if val is not None and len(val.keys()) != 0:
                fullLst = fullLst + list(val.values())
        for val in fullLst:
            retLst = retLst + list(val.values())
        return sorted(retLst, key=lambda val : val.number) if ordered else retLst



# global FULL_NAME
# global ONE_LETTER
from apalib.Data import data as data
global FLAGS
FLAGS = {}

#-w
#
class RNA:
    def __init__(self, seqNum=None, atoms=None, resName=None, set_name=True, chainID=None):
        self.SetResName(resName, set_name)
        self.SetSeqNum(seqNum)
        self.SetAtoms(atoms)
        self.SetChainID(chainID)

    def DeepCopy(self):
        return RNA(seqNum=self.seqNum,
                   atoms=self.atoms,
                   resName=self.resName,
                   set_name=True)

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetSeqNum(self, num):
        self.seqNum = num

    def SetAtoms(self, atoms):
        self.atoms = atoms
        self.CalculateCentroid(atoms)

    def GetAtoms(self):
        return self.atoms

    def CalculateCentroid(self, atoms):
        print("RNA CENTROID")

    def InsertAtom(self, atom):
        if self.atoms is None:
            self.atoms = list()
        self.atoms.append(atom)

    def SetResName(self, name, set_name):
        if not set_name:
            self.resName = name
            self.RaiseFlag('NO_NAME_CHECK')
            self.ClearFlag('BAD_NAME')
            return
        else:
            self.ClearFlag('NO_NAME_CHECK')
        global ONE_LETTER
        if data.ValidateRNA(name):
            self.resName = data.SetRNAName(name)
        # elif name in FULL_NAME:
        #     self.name = FULL_NAME[name]
        # else:
        #     self.RaiseFlag('BAD_NAME')
        #     return
        self.ClearFlag('BAD_NAME')
        return

    def GetResName(self):
        return self.resName

    def SetChainID(self, c):
        self.chainID = c

    def WriteForPDB(self):
        retstr = ""
        for atom in self.atoms:
            retstr += atom.WritePDB(intro="ATOM  ")
        return retstr

    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False

    def __lt__(self, other):
        return self.seqNum < other.number

    def __repr__(self):
        return f"RESIDUE: {self.resName}, NUMBER: {self.seqNum}"

    def __str__(self):
        return f"{self.resName} {self.seqNum}"

global ONE_LETTER
ONE_LETTER = {
    'A': 'ADENINE',
    'C': 'CYTOSINE',
    'G': 'GUANINE',
    'T': 'THYMINE',
    'U': 'URACIL',
    'I': 'INOSINE'
}
global FULL_NAME
FULL_NAME = {
    'ADENINE': 'A',
    'CYTOSINE': 'C',
    'GUANINE': 'G',
    'THYMINE': 'T',
    'URACIL': 'U',
    'INOSINE': 'I'
}



import apalib.config as config
import json
import apalib

pdb = apalib.PDB()

pdb.Fetch('5u59')

one = pdb.Contents().GetPeptideChains()['A'][1]
two = pdb.Contents().GetPeptideChains()['A'][2]

val = apalib.VectorPair(one, two)

print(apalib.GetCentAngle(one, two, rad=False))

data = apalib.Data()
pdb.Read(r"C:\Users\natem\OneDrive\Desktop\Research\ccSASA - Copy\1j1j\1j1j.mmol")


print("Stop")



# print("wow")
# #TODO Fix Mo
#
#
# class Element():
#     def __init__(self):
#         self.isotopes = []
#
#     def SetName(self, n):
#         self.name = n
#
#     def SetNumber(self, n):
#         self.number = n
#
#     def SetSymbol(self, s):
#         self.symbol = s
#
#     def SetIsotope(self, i, l):
#         self.isotopes.append([i, l[0], l[1]])
#         pass
#
# data = config.data.GetJson()
#
#
# with open("testjson.json","w") as fp:
#     json.dump(data, fp, indent=3)
#
# all = dict()
# # with open(r"C:\Users\natem\OneDrive\Desktop\final atoms.txt") as f:
# #     with open("output.txt", "w") as out:
# #         for line in f.read().splitlines():
# #             count = 0
# #             if line.find("Since") != -1:
# #                 out.write(line + "\n")
# #             else:
# #                 for i in range(len(line.split())):
# #                     if line.split()[i].find(''.join([i for i in line.split()[0] if not i.isdigit()])) != -1:
# #                         out.write('\n')
# #                     out.write(line.split()[i] + " ")
# #                 out.write('\n')
#
# with open(r'C:\Users\natem\PycharmProjects\apalib\output.txt') as f:
#     for line in f.read().splitlines():
#
#         if len(line.split()) == 0:
#             continue
#         if ''.join([i for i in line.split()[0] if not i.isdigit()]) not in all.keys():
#             all[''.join([i for i in line.split()[0] if not i.isdigit()])] = Element()
#         #New element
#         if line.find("Since") != -1:
#             number = ''.join([i for i in line.split()[0] if i.isdigit()])
#             symbol = ''.join([i for i in line.split()[0] if not i.isdigit()])
#             name = line.split()[1]
#             all[symbol].SetName(name)
#             all[symbol].SetSymbol(symbol)
#             all[symbol].SetNumber(number)
#         #Isotope of an element
#         else:
#             lst = line.split()
#             for i in range(len(lst)):
#                 lst[i] = lst[i].strip()
#             # print(lst)
#             symbol = ''.join([i for i in line.split()[0] if not i.isdigit()])
#             isotope = ''.join([i for i in line.split()[0] if i.isdigit()])
#             if lst[-1] == '1.0000':
#                 # print(line)
#                 pass
#             else:
#                 newlst = []
#                 flag = False
#                 for i in range(1, len(lst), 1):
#                     if flag:
#                         flag = False
#                         continue
#                     if lst[i].find(')') != -1:
#                         ln = lst[i - 1] + lst[i]
#                         newlst.append(ln)
#                         flag = True
#                 if len(newlst) == 1:
#                     newlst.append(lst[-1])
#                 for i in range(len(newlst)):
#                     newlst[i] = newlst[i][0:newlst[i].find('(')]
#
#                 try:
#                     all[symbol].SetIsotope(isotope, newlst)
#                 except IndexError:
#                     print(symbol, isotope, newlst)
#                 pass
#
#
# print("stop")
#
# for el in data['Atoms'].keys():
#     try:
#         # print(f"Comparing {data['Atoms'][el]['Symbol']}")
#         # print(data['Atoms'][el]['Symbol'], all[data['Atoms'][el]['Symbol']].symbol)
#         for iso in all[data['Atoms'][el]['Symbol']].isotopes:
#             flag = False
#             for iso2 in data['Atoms'][el]['Stable'].keys():
#                 print(iso[1], data['Atoms'][el]['Stable'][iso2]['Mass'])
#                 if iso[1] == str(data['Atoms'][el]['Stable'][iso2]['Mass']):
#                     flag = True
#             if not flag:
#                 print(data['Atoms'][el]['Symbol'], "Very bad")
#     except:
#         pass
# pass
# # for key in data['Atoms'].keys():
# #     print
# #     print(f"\"{key}\" : \"{key}\",")
# #     print(f"\"{data['Atoms'][key]['Symbol']}\" : \"{key}\",")
# #     # print("\"Isotopes\" : {}")
#
#
# # with open(r'C:\Users\natem\Downloads\IUPAC-atomic-masses.csv') as f:
# #     with open(r'C:\Users\natem\Downloads\IUPAC-2020.csv', 'w') as n:
# #         for line in f:
# #             if line.find(r'2020</a>') != -1:
# #                 n.write(line[:line.find('<')] + '\n')
# #
#
# # with open(r'C:\Users\natem\Downloads\IUPAC-2020.csv', 'r') as f:
# #     for line in f:
# #         line = line.replace(',',' ')
# #         spt = line.split()
# #         # print(spt[0])
# #         print(''.join([i for i in spt[0] if not i.isalpha()]), end=" ")
# #         print(config.data.Map('Atoms', ''.join([i for i in spt[0] if not i.isdigit()])), end=" ")
# #         print(''.join([i for i in spt[0] if not i.isdigit()]))
# #         if config.data.Map('Atoms', ''.join([i for i in spt[0] if not i.isdigit()])) not in info:
# #             info[config.data.Map('Atoms', ''.join([i for i in spt[0] if not i.isdigit()]))] = data['Atoms'][config.data.Map('Atoms', ''.join([i for i in spt[0] if not i.isdigit()]))]
# #         if ''.join([i for i in spt[0] if not i.isalpha()]) not in info[config.data.Map('Atoms', ''.join([i for i in spt[0] if not i.isdigit()]))]['Isotopes'].keys():
# #             info[config.data.Map('Atoms', ''.join([i for i in spt[0] if not i.isdigit()]))]['Isotopes'][''.join([i for i in spt[0] if not i.isalpha()])] = dict()
# #             info[config.data.Map('Atoms', ''.join([i for i in spt[0] if not i.isdigit()]))]['Isotopes'][
# #                 ''.join([i for i in spt[0] if not i.isalpha()])]['Mass'] = 0
# #             info[config.data.Map('Atoms', ''.join([i for i in spt[0] if not i.isdigit()]))]['Isotopes'][
# #                 ''.join([i for i in spt[0] if not i.isalpha()])]['Abundance'] = 0
#
#     # "Hydrogen":{
#     #   "Symbol": "H",
#     #   "Number": 1,
#     #   "Isotopes": {}
#     # },
#
# # print(info)
# # apalib.pdb.Fetch('5u59')
# # dic = config.Container.GetPeptideChains()
# # val = config.Container.AsList()
# #
# # val.sort()
# #
# # com = (dic['A'][9].GetRotamers(unique=True))
# #
# #

# ########################################################################################################################
# PrintSettings()
# Fetch('3mzw')
# Parse()
# print(GetProteinChains()['A'][1])
# print(getIEP(7.4))
# print(GetProteinChains()['A'][1].GetAtoms()[0].GetCoordinates(),
# #GetProteinChains()['A'][1].GetAtoms()[2].GetCoordinates())
# print(GetDistance(GetProteinChains()['A'][1].GetAtoms()[0].GetCoordinates(),
# #GetProteinChains()['A'][1].GetAtoms()[2].GetCoordinates()))


import apalib.config
from apalib.Data import data as data

global FLAGS
FLAGS = {}

class HETATM:
    def __init__(self, seqNum=None, atoms=None, resName=None, chainID=None):
        self.SetResName(resName)
        self.SetSeqNum(seqNum)
        self.SetAtoms(atoms)
        self.SetChainID(chainID)

    def DeepCopy(self):
        return HETATM(seqNum=self.seqNum,
                      atoms=self.atoms,
                      resName=self.resName)

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetSeqNum(self, num):
        self.seqNum = num

    def GetAtoms(self):
        return self.atoms

    def SetAtoms(self, atoms):
        self.atoms = atoms
        # self.CalculateCentroid()

    def InsertAtom(self, atom):
        if self.atoms is None:
            self.atoms = list()
        self.atoms.append(atom)

    def SetResName(self, resName):
        self.resName = resName

    def GetResName(self):
        return self.resName

    def ClearFlags(self):
        self.flags.clear()

    def CalculateCentroid(self):
        if 'atoms' not in self.__dict__:
            self.centroid = None
            return
        n = 0
        x = 0
        y = 0
        z = 0
        for atom in self.atoms:
            n += 1
            x += atom.GetCoordinates()[0]
            y += atom.GetCoordinates()[1]
            z += atom.GetCoordinates()[2]
        self.centroid = [x/n, y/n, z/n]

    def GetCentroid(self):
        if 'centroid' in self.__dict__:
            return self.centroid
        return None

    def SetChainID(self, c):
        self.chainID = c

    def WriteForPDB(self):
        retstr = ""
        for atom in self.atoms:
            retstr += atom.WritePDB(intro="HETATM")
        return retstr


    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False

    def __lt__(self, other):
        return self.seqNum < other.seqNum

    def __repr__(self):
        return f"RESIDUE: {self.resName}, seqNum: {self.seqNum}"

    def __str__(self):
        return f"{self.resName} {self.seqNum}"



from apalib.Container import Container as Container
# import apalib.Container
# from apalib.AminoAcid import AminoAcid
# from apalib.Atom import Atom
# from apalib.DNA import DNA
# from apalib.RNA import RNA
# from apalib.HETATM import HETATM
from apalib.Data import Data
Container = Container()
data = Data()


def ToStringLen(val, l, left_justify = True):
    if val is None:
        val = ''
    retstr = str(val)
    if len(retstr) < l and left_justify:
        return retstr + " " * (l - len(retstr))
    elif len(retstr) < l and not left_justify:
        return " " * (l - len(retstr)) + retstr
    else:
        return retstr


import sys
from apalib import Container
import apalib.config
from apalib.Data import data as data
global FLAGS
FLAGS = {}


class Atom:
    def __init__(self, serial = None, name = None, altLoc = None, resName = None, chainID = None, resSeq = None,
                 iCode = None, x = None, y = None, z = None, occupancy = None, tempFactor = None, element = None,
                 charge = None, extract_missing_element_from_name = True):
        self.SetSerial(serial)
        self.SetName(name, extract_missing_element_from_name)
        self.SetAltLoc(altLoc)
        self.SetResName(resName)
        self.SetChainID(chainID)
        self.SetResSeq(resSeq)
        self.SetICode(iCode)
        self.SetXYZ(x,y,z)
        self.SetOccupancy(occupancy)
        self.SetTempFactor(tempFactor)
        if element is not None and element != '':
            self.SetElement(element)
        self.SetCharge(charge)
        for key in self.__dict__:
            if self.__dict__[key] == '':
                self.__dict__[key] = None

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetSerial(self, serial):
        self.serial = serial

    def GetSerial(self):
        return self.serial

    def SetName(self, name, extract):
        self.name = name
        if extract:
            self.__ExtractElement(name)

    def GetName(self):
        return self.name

    def SetAltLoc(self, altLoc):
        self.altLoc = altLoc

    def GetAltLoc(self):
        return self.altLoc

    def SetResName(self, resName):
        self.resName = resName

    def GetResName(self):
        return self.resName

    def SetChainID(self, chainID):
        self.chainID = chainID

    def GetChainID(self):
        return self.chainID

    def SetResSeq(self, resSeq):
        self.resSeq = resSeq

    def GetResSeq(self):
        return self.resSeq

    def SetICode(self, iCode):
        self.iCode = iCode

    def GetICode(self):
        return self.iCode

    def SetXYZ(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z
        self.SetCoordinates([self.x, self.y, self.z])

    def GetXYZ(self):
        return [self.x, self.y, self.z]

    # def SetY(self, y):
    #     self.y = y
    #     self.coordinates[1] = y
    #
    # def GetY(self):
    #     return self.y
    #
    # def SetZ(self, z):
    #     self.z = z
    #     self.coordinates[2] = z
    #
    # def GetZ(self):
    #     return self.z

    def SetCoordinates(self, coordinates):
        self.coordinates = [coordinates[0], coordinates[1], coordinates[2]]
        if isinstance(coordinates, list) and len(coordinates) == 3:
            self.coordinates = [float(x.strip()) for x in self.coordinates]

    def GetCoordinates(self):
        return self.coordinates

    def SetOccupancy(self, occupancy):
        self.occupancy = occupancy

    def GetOccupancy(self):
        return self.occupancy

    def SetTempFactor(self, tempFactor):
        self.tempFactor = tempFactor

    def GetTempFactor(self):
        return self.tempFactor

    def SetElement(self, element):
        self.element = element

    def GetElement(self):
        return self.element

    def SetCharge(self, charge):
        self.charge = charge

    def GetCharge(self):
        return self.charge

    # def SetCoordinates(self, coor):
    #     self.coordinates = [float(coor[0]), float(coor[1]), float(coor[2])]
    #
    # def GetCoordinates(self):
    #     if 'coordinates' in self.__dict__:
    #         return self.coordinates
    #     return None
    #
    # def SetID(self, id, extract):
    #     self.id = id
    #     if extract:
    #         self.__ExtractElement(id)
    #
    # def GetID(self):
    #     if 'id' in self.__dict__:
    #         return self.id
    #     return None
    #
    # def SetOccupancy(self, occ):
    #     self.occupancy = occ
    #
    # def GetOccupancy(self):
    #     if 'occupancy' in self.__dict__:
    #         return self.occupancy
    #     return None
    #
    # def SetBFactor(self, bfact):
    #     self.b_factor = bfact
    #
    # def GetBFactor(self):
    #     if 'b_factor' in self.__dict__:
    #         return self.b_factor
    #     return None
    #
    # def SetElement(self, ele):
    #     if ele != '':
    #         self.element = ele
    #
    # def GetElement(self):
    #     if 'element' in self.__dict__:
    #         return self.element
    #     return None
    #
    # def SetResidue(self, res):
    #     #TODO Set rotation based off of parameter. Add a bool parameter?
    #     if len(res) == 4:
    #         self.rotation = res[0]
    #         self.residue = res[1:]
    #         return
    #     else:
    #         self.rotation = ""
    #     self.residue = res
    #
    # def GetResidue(self):
    #     if 'residue' in self.__dict__:
    #         return self.residue
    #     return None

    #In case the element informatin is missing
    def __ExtractElement(self, id):
        if 'CU' in id:
            self.SetElement('CU')
        elif 'SE' in id:
            self.SetElement('SE')
        elif 'O' in id:
            self.SetElement('O')
        elif 'ZN' in id:
            self.SetElement('ZN')
        elif 'P' in id:
            self.SetElement('P')
        elif 'N' in id:
            self.SetElement('N')
        elif 'C' in id:
            self.SetElement('C')
        elif 'S' in id:
            self.SetElement('S')
        elif 'MG' in id:
            self.SetElement('MG')
        elif 'H' in id:
            self.SetElement('H')
        elif "FE" in id.upper():
            self.SetElement('FE')
        else:
            sys.exit("SOMETHING WENT WRONG! CHECK WHAT HAPPENED! THIS ERROR SHOULD NOT OCCUR")

    def GetRadius(self, ver = "Van der Waals"):
        name = data.Map('Elements', self.element)
        return data.GetJson()['Atoms'][name]['Radius'][ver]

    def GetDistance(self, other):
        return ((self.coordinates[0] - other.coordinates[0])**2 + (self.coordinates[1] - other.coordinates[1])**2 + (self.coordinates[2] - other.coordinates[2])**2)**.5

    def GetConfiguration(self, extended=False):
        name = data.Map('Elements', self.element)
        if not extended:
            return data.GetJson()['Atoms'][name]["Ground Configuration"]
        else:
            return data._BuildConfiguration(name)

    def WritePDB(self, intro="ATOM  "):
        retstr = intro
        retstr += apalib.config.ToStringLen(self.serial, 5, left_justify=False)
        retstr += "  "
        retstr += apalib.config.ToStringLen(self.name, 3, left_justify=True)
        retstr += apalib.config.ToStringLen(self.altLoc, 1, left_justify=False)
        retstr += apalib.config.ToStringLen(self.resName, 3, left_justify=False)
        retstr += " "
        retstr += apalib.config.ToStringLen(self.chainID, 1, left_justify=False)
        retstr += apalib.config.ToStringLen(self.resSeq, 4, left_justify=False)
        retstr += " "  # TODO Verify this. What is AChar?
        retstr += "   "
        retstr += apalib.config.ToStringLen(self.x, 8, left_justify=False)
        retstr += apalib.config.ToStringLen(self.y, 8, left_justify=False)
        retstr += apalib.config.ToStringLen(self.z, 8, left_justify=False)
        retstr += apalib.config.ToStringLen(self.occupancy, 6, left_justify=False)
        retstr += apalib.config.ToStringLen(self.tempFactor, 6, left_justify=False)
        retstr += apalib.config.ToStringLen("", 8, left_justify=True)
        retstr += apalib.config.ToStringLen(self.element, 4, left_justify=False)
        retstr += apalib.config.ToStringLen(self.charge, 2, left_justify=True)
        retstr += "\n"
        return retstr

    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False


    #TODO Make this not crash if an element is missing. Probably iterate through self.__dict__()
    def __repr__(self):
        retStr = "Atom: "
        for key in self.__dict__:
            retStr += f"{key}: {self.__dict__[key]}, "
        return retStr

        # TODO Create some setting where this will print instead
        return f"ATOM: NUMBER: {self.number}, " \
               f"TAG: {self.id}, RESIDUE: {self.residue}, " \
               f"COORDINATES: {self.coordinates}, OCCUPANCY: {self.occupancy}," \
               f" B_FACTOR: {self.bfactor}, ELEMENT: {self.element}"

    def __str__(self):
        retStr = ""
        for key in self.__dict__:
            retStr += f"{key}: {self.__dict__[key]}, "
        return retStr

        # TODO Create some setting where this will print instead
        return f"{self.id} {self.rotation}{self.residue} {self.coordinates}"


from unittest import TestCase


class TestAtom(TestCase):
    def test_set_number(self):
        self.fail()

    def test_set_coordinates(self):
        self.fail()

    def test_get_coordinates(self):
        self.fail()

    def test_set_id(self):
        self.fail()

    def test_set_occupancy(self):
        self.fail()

    def test_set_bfactor(self):
        self.fail()

    def test_set_element(self):
        self.fail()

    def test_set_residue(self):
        self.fail()



class NoFetchError(Exception):
    #Used when a fetched pdb file is required but not provided
    def __init__(self):
        self.message = 'No fetched protein. Use method Fetch(PDB_CODE) to fetch a protein.\n'
        return self.message

    def __str__(self):
        return self.message

class BadInternet(Exception):
    def __init__(self):
        self.message = 'Poor or no internet connection. Some functionality may be lost\n\t' \
                       r'Connection to http://www.google.com timeout or failure'

    def __str__(self):
        return self.message

class BadKwarg(Exception):
    def __init__(self, accepted):
        self.message = f'Bad parameter provided. Accepted parameters are as follows: {accepted}'

    def __str__(self):
        return self.message


global FLAGS
FLAGS = {}

class HETATM:
    def __init__(self, **kwargs):
        self.number = None
        self.atoms = None
        self.name = None

        if 'name' in kwargs:
            self.SetName(kwargs['name'])
        if 'number' in kwargs:
            self.SetNumber(kwargs['number'])
        if 'atoms' in kwargs:
            self.SetAtoms(kwargs['atoms'])

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetNumber(self, num):
        self.number = num

    def SetAtoms(self, atoms):
        self.atoms = atoms
        self.CalculateCentroid(atoms)

    def InsertAtom(self, atom):
        if self.atoms is None:
            self.atoms = list()
        self.atoms.append(atom)

    def SetName(self, name):
        self.name = name

    def ClearFlags(self):
        self.flags.clear()

    def CalculateCentroid(self, atoms):
        n = 0
        x = 0
        y = 0
        z = 0
        for atom in atoms:
            n += 1
            x += atom.GetCoordinates()[0]
            y += atom.GetCoordinates()[1]
            z += atom.GetCoordinates()[2]
        self.centroid = [x/n, y/n, z/n]

    def GetCentroid(self):
        if 'centroid' in self.__dict:
            return self.centroid
        return None

    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False

    def __lt__(self, other):
        return self.number < other.number

    def __repr__(self):
        return f"RESIDUE: {self.name}, NUMBER: {self.number}"

    def __str__(self):
        return f"{self.name} {self.number}"



import sys
global FLAGS
FLAGS = {}

class Atom:
    def __init__(self, **kwargs):
        extract = True
        if 'extract' in kwargs:
            extract = kwargs['extract']
        if 'number' in kwargs:
            self.SetNumber(kwargs['number'])
        if 'coordinates' in kwargs:
            self.SetCoordinates(kwargs['coordinates'])
        if 'id' in kwargs:
            self.SetID(kwargs['id'], extract)
        if 'occupancy' in kwargs:
            self.SetOccupancy(kwargs['occupancy'])
        if 'b_factor' in kwargs:
            self.SetBFactor(kwargs['b_factor'])
        if 'element' in kwargs:
            self.SetElement(kwargs['element'])
        if 'residue' in kwargs:
            self.SetResidue(kwargs['residue'])

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetNumber(self, num):
        self.number = num

    def GetNumber(self):
        if 'number' in self.__dict__:
            return self.number
        return None

    def SetCoordinates(self, coor):
        self.coordinates = [float(coor[0]), float(coor[1]), float(coor[2])]

    def GetCoordinates(self):
        if 'coordinates' in self.__dict__:
            return self.coordinates
        return None

    def SetID(self, id, extract):
        self.id = id
        if extract:
            self.__ExtractElement(id)

    def GetID(self):
        if 'id' in self.__dict__:
            return self.id
        return None

    def SetOccupancy(self, occ):
        self.occupancy = occ

    def GetOccupancy(self):
        if 'occupancy' in self.__dict__:
            return self.occupancy
        return None

    def SetBFactor(self, bfact):
        self.b_factor = bfact

    def GetBFactor(self):
        if 'b_factor' in self.__dict__:
            return self.b_factor
        return None

    def SetElement(self, ele):
        self.element = ele

    def GetElement(self):
        if 'element' in self.__dict__:
            return self.element
        return None

    def SetResidue(self, res):
        #TODO Set rotation based off of parameter. Add a bool parameter?
        if len(res) == 4:
            self.rotation = res[0]
            self.residue = res[1:]
            return
        else:
            self.rotation = ""
        self.residue = res

    def GetResidue(self):
        if 'residue' in self.__dict__:
            return self.residue
        return None

    #In case the element informatin is missing
    def __ExtractElement(self, id):
        if 'C' in id:
            self.SetElement('C')
        elif 'S' in id:
            self.SetElement('S')
        elif 'O' in id:
            self.SetElement('O')
        elif 'N' in id:
            self.SetElement('N')
        elif 'P' in id:
            self.SetElement('P')
        elif 'ZN' in id:
            self.SetElement('ZN')
        elif 'CU' in id:
            self.SetElement('CU')
        elif 'SE' in id:
            self.SetElement('SE')
        elif 'MG' in id:
            self.SetElement('MG')
        elif 'H' in id:
            self.SetElement('H')
        else:
            sys.exit("SOMETHING WENT WRONG! CHECK WHAT HAPPENED! THIS ERROR SHOULD NOT OCCUR")

    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False


    #TODO Make this not crash if an element is missing. Probably iterate through self.__dict__()
    def __repr__(self):
        retStr = "Atom: "
        for key in self.__dict__:
            retStr += f"{key}: {self.__dict__[key]}, "
        return retStr

        # TODO Create some setting where this will print instead
        return f"ATOM: NUMBER: {self.number}, " \
               f"TAG: {self.id}, RESIDUE: {self.residue}, " \
               f"COORDINATES: {self.coordinates}, OCCUPANCY: {self.occupancy}," \
               f" B_FACTOR: {self.bfactor}, ELEMENT: {self.element}"

    def __str__(self):
        retStr = ""
        for key in self.__dict__:
            retStr += f"{key}: {self.__dict__[key]}, "
        return retStr

        # TODO Create some setting where this will print instead
        return f"{self.id} {self.rotation}{self.residue} {self.coordinates}"


from apalib.AminoAcid import AminoAcid
from apalib.Atom import Atom
from apalib.DNA import DNA
from apalib.RNA import RNA
from apalib.HETATM import HETATM
from apalib.Data import data as data

class Container:
    def __init__(self):
        self.current_fetch = None
        self.PeptideChains = None
        self.DNAChains = None
        self.RNAChains = None
        self.HETATMChains = None

    def ClearAll(self, initialize = True):
        self.__init__()
        if initialize:
            self.HETATMChains = {}
            self.PeptideChains = {}
            self.DNAChains = {}
            self.RNAChains = {}

    def AddChain(self, key):
        self.HETATMChains[key] = {}
        self.PeptideChains[key] = {}
        self.DNAChains[key] = {}
        self.RNAChains[key] = {}

    #Create a new residue if missing. Otherwise return the designated residue
    def AddResidue(self, resType, index, resName, chainID):
        if chainID not in self.HETATMChains.keys():
            self.AddChain(chainID)
        index = int(index.strip())
        if resType == 'HETATM':
            if index in self.HETATMChains[chainID].keys():
                return self.HETATMChains[chainID][index]
            self.HETATMChains[chainID][index] = HETATM(seqNum=index, resName=resName, chainID=chainID)
            return self.HETATMChains[chainID][index]

        elif resType == 'RNA':
            if index in self.RNAChains[chainID].keys():
                return self.RNAChains[chainID][index]
            self.RNAChains[chainID][index] = RNA(seqNum=index, resName=resName, chainID=chainID)
            return self.RNAChains[chainID][index]

        elif resType == 'DNA':
            if index in self.DNAChains[chainID].keys():
                return self.DNAChains[chainID][index]
            self.DNAChains[chainID][index] = DNA(seqNum=index, resName=resName, chainID=chainID)
            return self.DNAChains[chainID][index]

        elif resType == 'AA':
            if index in self.PeptideChains[chainID].keys():
                return self.PeptideChains[chainID][index]
            self.PeptideChains[chainID][index] = AminoAcid(seqNum=index, resName=resName, chainID=chainID)
            return self.PeptideChains[chainID][index]
        else:
            import sys
            sys.stderr.write("Something went wrong in adding a residue")


    def SetFetch(self, fetch):
        self.current_fetch = fetch

    def GetFetch(self):
        return self.current_fetch

    def ClearFetch(self):
        self.current_fetch = None

    def SetProteinChains(self, pchain):
        self.PeptideChains = pchain

    def GetPeptideChains(self):
        return self.PeptideChains

    def ClearPeptideChains(self):
        self.PeptideChains = None

    def SetDNAChains(self,dchain):
        self.DNAChains = dchain

    def GetDNAChains(self):
        return self.DNAChains

    def ClearDNAChains(self):
        self.DNAChains = None

    def GetRNAChains(self):
        return self.RNAChains

    def SetRNAChains(self, rchain):
        self.RNAChains = rchain

    def ClearRNAChains(self):
        self.RNAChains = None

    def GetHETATMChains(self):
        return self.HETATMChains

    def SetHETATMChains(self, hchain):
        self.HETATMChains = hchain

    def ClearHEETATMChains(self):
        self.HETATMChains = None

    #Perform calculations, etc. to be done after a full parse
    def _PostParseEvaluations(self):
        #Fill in variables
        for chain in self.PeptideChains.keys():
            for res in self.PeptideChains[chain].values():
                res.CalculateCentroid()
        pass



    #Returns all residues from all children of CONTAINER as a list
    #Nasty list comprehension goes faster than a for-loop
    def DumpResidues(self):
        return [val for sublist in (list(res.values()) for res in list(self.PeptideChains.values()) + list(self.HETATMChains.values()) + list(self.DNAChains.values()) + list(self.RNAChains.values())) for val in sublist]
        # lst =  list(self.PeptideChains.values()) + list(self.HETATMChains.values()) + list(self.DNAChains.values()) + list(self.RNAChains.values())
        # retlst = []
        # for d in lst:
        #     retlst += list(d.values())
        # return retlst

    #Return all atoms from all children of CONTAINER as a list
    #Virtually human-unreadable, but its efficient and pythonic
    def DumpAtoms(self):
        return [atom for sublist in (group.GetAtoms() for group in (val for sublist in (list(res.values()) for res in list(self.PeptideChains.values()) + list(self.HETATMChains.values()) + list(self.DNAChains.values()) + list(self.RNAChains.values())) for val in sublist)) for atom in sublist]
        # return [atom for sublist in (group.GetAtoms() for group in self.DumpResidues()) for atom in sublist]




        # retLst = []
        # val = self.DumpResidues()
        # for key in self.PeptideChains.keys():
        #     print(self.PeptideChains.values())
        #     # retLst += self.PeptideChains[key].GetAtoms()
        # for key in self.HETATMChains.keys():
        #     pass
        #     # retLst += self.HETATMChains[key].GetAtoms()
        # for key in self.DNAChains.keys():
        #     pass
        #     # retLst += self.DNAChains[key].GetAtoms()
        # for key in self.RNAChains.keys():
        #     pass
        #     # retLst += self.RNAChains[key].GetAtoms()
        # return retLst

    #Return all residues from all chains as a single list
    def AsList(self, ordered=True):
        fullLst = []
        retLst = []
        lst = [self.PeptideChains, self.DNAChains, self.RNAChains, self.HETATMChains]
        for val in lst:
            if val is not None and len(val.keys()) != 0:
                fullLst = fullLst + list(val.values())
        for val in fullLst:
            retLst = retLst + list(val.values())
        return sorted(retLst, key=lambda val : val.seqNum) if ordered else retLst



import apalib.apalibExceptions
import apalib.apalibExceptions as apaExcept
from apalib.Atom import Atom
from apalib.HETATM import *
from apalib.AminoAcid import *
import sys
from apalib import *
from apalib.Data import data as data

class PDB:
    def __init__(self):
        self.container = Container()
        # print(apalib.j_data.GetJson())

    def Contents(self):
        return self.container

    def FetchFASTA(self, prot):
        import urllib.request
        url = r"https://www.rcsb.org/fasta/entry/" + prot.strip().upper() + r"/display"
        try:
            with urllib.request.urlopen(url) as f:
                return f.read().decode('utf-8')
        except urllib.error.URLError:
            sys.stderr.write("The requested pdb code could not be retrieved or does not exist\n")
            return False

    # def FetchAsFile(self, prot):
    #     import urllib.request
    #     url = r'https://files.rcsb.org/download/' + prot.strip() + '.pdb'
    #     try:
    #         with urllib.request.urlopen(url) as f:


    def Fetch(self, prot, crash = True, hold_pdb=False):
        # print("Fetching ", prot)
        import urllib.request
        url = r'https://files.rcsb.org/download/' + prot.strip() + '.pdb'
        try:
            with urllib.request.urlopen(url) as f:
                self.container.SetFetch(f.read().decode('utf-8'))
                self._Parse(hold_pdb)
                return True
        except urllib.error.URLError:
            sys.stderr.write("The requested pdb code could not be retrieved or does not exist\n")
            if crash:
                exit()
            return False

    def Read(self, path, hold_pdb=False):
        with open(path, 'r') as fp:
            self.container.SetFetch(fp.read())
            self._Parse(hold_pdb)

    # Wrapper for the ParsePDB file to allow functionality with a fetched protein
    def _Parse(self, hold_pdb=False):
        try:
            if self.container.GetFetch() is None:
                raise apaExcept.NoFetchError
            return self._ParsePDB(self.container.GetFetch(), hold_pdb)
            # return self._ParsePDB(self.container.GetFetch().splitlines())
        except apaExcept.NoFetchError as e:
            sys.stderr.write(e.message)


    #PDB standard described here: https://www.wwpdb.org/documentation/file-format-content/format33/v3.3.html
    def _ParsePDB(self, raw_pdb, hold_pdb=False):
        self.container.ClearAll()
        if hold_pdb:
            self.container.SetFetch(raw_pdb)
        for line in raw_pdb.splitlines():
            if line[0:6] == 'ATOM  ' or line[0:6] == 'HETATM':
                self._ParseAtomHETATM(line)
            #TODO Parse REMARK 350 to get symmetry information

        self.container._PostParseEvaluations()

    def _ParseAtomHETATM(self, line):
        serial = line[6:11].strip()
        name = line[12:16].strip()
        altLoc = line[16].strip()
        resName = line[17:20].strip()
        chainID = line[21].strip()
        resSeq = line[22:26].strip()
        iCode = line[26].strip()
        x = line[30:38].strip()
        y = line[38:46].strip()
        z = line[46:54].strip()
        occupancy = line[54:60].strip()
        tempFactor = line[60:66].strip()
        element = line[76:78].strip()
        charge = line[78:80].strip()
        atom = Atom.Atom(serial=serial, name=name, altLoc=altLoc, resName=resName, chainID=chainID, resSeq=resSeq,
                         iCode=iCode, x=x, y=y, z=z, occupancy=occupancy, tempFactor=tempFactor, element=element,
                         charge=charge)
        resType = self.DetermineResType(resName)
        residue = self.container.AddResidue(resType, resSeq, resName, chainID)
        residue.InsertAtom(atom)

    def DetermineResType(self, resName):
        if data.ValidateRNA(resName):
            return 'RNA'
        elif data.ValidateDNA(resName):
            return 'DNA'
        elif data.ValidateAA(resName):
            return "AA"
        else:
            return "HETATM"

    #Remove all of the waters from the current fetch. Probably make this more general for any HETATM group. Make a wrapper?
    def RemoveWater(self):
        h_chains = self.container.GetHETATMChains()
        for chain in h_chains.keys():
            h_chains[chain] = {key: value for (key, value) in h_chains[chain].items() if value.GetResName().upper() != 'HOH'}

    # def Validate(self, **kwargs):
    #     for key in kwargs:
    #         if key != 'pdb' or (key == 'pdb' and not isinstance(kwargs['pdb'], str)):
    #             raise apalib.apalibExceptions.BadKwarg('pdb=<pdb_to_validate>')

    #Write contents to a PDB file
    def WritePDB(self, fp):
        wr = ""
        s = sorted(self.container.DumpResidues(), key=lambda x: x.seqNum)
        with open(fp, "w") as f:
            for res in s:
                f.write(res.WriteForPDB())

    #Write contents to FASTA
    def ToFASTA(self):
        ls = self.container.AsList(ordered=True)
        retStr = ""
        for r in ls:
            if data.ValidateAA(r.resName):
                name = data.Map("Amino Acids", r.resName)
                retStr += data.GetJson()["Amino Acids"][name]["1code"]
            elif r.resName.upper() != "HOH":
                retStr += "X"
        return retStr


import sys
from apalib.config import *
global FLAGS
FLAGS = {}


class AminoAcid:
    def __init__(self, **kwargs):
        self.number = None
        self.atoms = None
        self.name = None
        self.rotamer = None
        self.vector = None
        self.heptad = None
        self.centroid = None
        set_name = True
        if 'set_name' in kwargs:
            set_name = kwargs['set_name']
        if 'name' in kwargs:
            self.SetName(kwargs['name'], set_name)
        if 'number' in kwargs:
            self.SetNumber(kwargs['number'])
        if 'atoms' in kwargs:
            self.SetAtoms(kwargs['atoms'])

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetNumber(self, num):
        self.number = num

    def SetAtoms(self, atoms):
        self.atoms = atoms
        self.CalculateCentroid(atoms)

    def InsertAtom(self, atom):
        if self.atoms is None:
            self.atoms = list()
        self.atoms.append(atom)

    def GetAtoms(self):
        return self.atoms

    def GetRotamers(self, **kwargs):
        accepted = ['unique']
        for key in kwargs.keys():
            if key not in accepted:
                raise apalib.apalibExceptions.BadKwarg(accepted)

        if self.rotamer is not None:
            return self.rotamer

        if 'unique' in kwargs.keys() and isinstance(kwargs['unique'], (bool)):
            unique = kwargs['unique']
        else:
            unique = False

        retDict = {'Common':[]}
        if self.atoms is not None:
            for atom in self.atoms:
                if atom.rotation is None or atom.rotation == '':
                    retDict['Common'].append(atom)
                    continue
                if atom.rotation not in retDict.keys():
                    retDict[atom.rotation] = []
                retDict[atom.rotation].append(atom)
        if unique is False:
            for lst in [retDict[key] for key in retDict.keys() if key != 'Common']:
                lst += retDict['Common']
        return retDict

    def GetCA(self):
        if self.atoms is None:
            return None
        for atom in self.atoms:
            if atom.GetID() == 'CA':
                return atom
        return None

    def SetName(self, name, set_name):
        if not set_name:
            self.name = name
            self.RaiseFlag('NO_NAME_CHECK')
            return
        else:
            self.ClearFlag('NO_NAME_CHECK')
        if len(name) == 3 and name in THREE_LETTER:
            self.name = name
        elif len(name) == 3 and name not in THREE_LETTER:
            self.name = None
        elif len(name) == 1:
            self.name = self.OneToThree(name)
        elif len(name) <= 4:
            if name[-3:] in THREE_LETTER:
                self.name = name[-3:]
                self.rotamer = name[:-3]
            else:
                self.name = None
        elif len(name) == 2:
            self.name = None
        if self.name is not None:
            self.RaiseFlag('BAD_NAME')
        else:
            self.RaiseFlag('BAD_NAME')
        if self.name is not None and self.rotamer is not None:
            self.RaiseFlag('MARKED')
        elif self.name is not None and self.rotamer is None:
            self.ClearFlag('MARKED')

    def SetHeptad(self, heptad):
        self.heptad = heptad

    def CalculateCentroid(self, atoms):
        #TODO switch to using the JSON for this
        # *For Glycine, only the alpha carbon is considered
        # *For Alanine, only the beta carbon is considered
        AAs = {
            "SER": ['OG'],
            "CYS": ['SG'],
            "SEC": ['SE'],
            "GLY": ['CA'],
            "ALA": ['CB'],
            "THR": ['OG1', 'CG2'],
            "PRO": ['CG', 'CD'],
            "VAL": ['CG1', 'CG2'],
            "ASP": ['CG', 'OD1', 'OD2'],
            "ASN": ['CG', 'OD1', 'ND2'],
            "ILE": ['CG1', 'CG2', 'CD1'],
            "LEU": ['CG', 'CD1', 'CD2'],
            "MET": ['CG', 'SD', 'CE'],
            "LYS": ['CG', 'CD', 'CE', 'NZ'],
            "GLU": ['CG', 'CD', 'OE1', 'OE2'],
            "GLN": ['CG', 'CD', 'OE1', 'NE2'],
            "HIS": ['CG', 'ND1', 'CD2', 'CE1', 'NE2'],
            "ARG": ['CG', 'CD', 'NE', 'CZ', 'NH1', 'NH2'],
            "PHE": ['CG', 'CD1', 'CD2', 'CE1', 'CE2', 'CZ'],
            "TYR": ['CG', 'CD1', 'CD2', 'CE1', 'CE2', 'CZ', 'OH'],
            "TRP": ['CG', 'CD1', 'CD2', 'NE1', 'CE2', 'CE3', 'CZ2', 'CZ3', 'CH2']
        }
        # TODO Deal with non-1 occupancy
        if len(self.name) == 3 and self.name in AAs:
            # If residue is not specified
            x_coord = 0
            y_coord = 0
            z_coord = 0
            num_atoms = 0
            for atom in [atom for atom in atoms if atom.id in AAs[self.name]]:
                if atom.element == 'H':
                    continue
                x_coord += atom.GetCoordinates()[0]
                y_coord += atom.GetCoordinates()[1]
                z_coord += atom.GetCoordinates()[2]
                num_atoms += 1
            self.centroid = list()
            try:
                self.centroid.append(x_coord / num_atoms)
                self.centroid.append(y_coord / num_atoms)
                self.centroid.append(z_coord / num_atoms)
            except ZeroDivisionError:
                self.centroid.clear()
                beta = [atom for atom in atoms if atom.id == 'CB']
                alpha = [atom for atom in atoms if atom.id == 'CA']
                if len(beta) != 0:
                    self.centroid.append(beta[0].coordinates[0])
                    self.centroid.append(beta[0].coordinates[1])
                    self.centroid.append(beta[0].coordinates[2])
                    self.RaiseFlag('B_CENTROID')
                elif len(alpha) != 0:
                    self.centroid.append(alpha[0].coordinates[0])
                    self.centroid.append(alpha[0].coordinates[1])
                    self.centroid.append(alpha[0].coordinates[2])
                    self.RaiseFlag('A_CENTROID')
                else:
                    self.centroid = None
                    self.vector = None
                    self.RaiseFlag('BAD_CENTROID')
                    return
            # After all that, set the centroidal vector
            self.vector = [self.centroid[0] - self.GetCA().GetCoordinates()[0],
                           self.centroid[1] - self.GetCA().GetCoordinates()[1],
                           self.centroid[2] - self.GetCA().GetCoordinates()[2]]

        elif len(self.name) == 4 and self.name[1:] in AAs:
            sys.stderr.write("ROTAMER!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        else:
            sys.exit("UNKNOWN AMINO ACID YO. DO SOMETHING ABOUT THIS!!!!!")

    def GetCentroid(self):
        return self.centroid

    def GetVector(self):
        return self.vector

    @staticmethod
    def OneToThree(oln):
        if oln in ONE_LETTER:
            try:
                return ONE_LETTER[oln]
            except KeyError:
                return None

    @staticmethod
    def NameToThree(name):
        if name in FULL_NAME:
            try:
                return FULL_NAME[name]
            except KeyError:
                return None

    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False

    # TODO These would be pretty cool to implement if possible. Open and write to a new python file that these access?
    @staticmethod
    def Set_lt(str):
        print('stub')

    @staticmethod
    def Set_repr(str):
        print('stub')


    def Set_str(self, str):
        return

    def __lt__(self, other):
        return self.number < other.number

    def __repr__(self):
        return f"RESIDUE: {self.name}, NUMBER: {self.number}"

    def __str__(self):
        return f"{self.name} {self.number}"


# Shoved down here for cleanliness
global ONE_LETTER
ONE_LETTER = {
    "R": 'ARG',
    "H": 'HIS',
    "K": 'LYS',
    "D": 'ASP',
    "E": 'GLU',
    "S": 'SER',
    "T": 'THR',
    "N": 'ASN',
    "Q": 'GLN',
    "C": 'CYS',
    "U": 'SEC',
    "G": 'GLY',
    "P": 'PRO',
    "A": 'ALA',
    "V": 'VAL',
    "I": 'ILE',
    "L": 'LEU',
    "M": 'MET',
    "F": 'PHE',
    "Y": 'TYR',
    "W": 'TRP',
    "O": 'PYL'
}

global THREE_LETTER
THREE_LETTER = {
    "ARG": 'R',
    "HIS": 'H',
    "LYS": 'K',
    "ASP": 'D',
    "GLU": 'E',
    "SER": 'S',
    "THR": 'T',
    "ASN": 'N',
    "GLN": 'G',
    "CYS": 'C',
    "SEC": 'U',
    "GLY": 'G',
    "PRO": 'P',
    "ALA": 'A',
    "VAL": 'V',
    "ILE": 'I',
    "LEU": 'L',
    "MET": 'M',
    "PHE": 'F',
    "TYR": 'Y',
    "TRP": 'W',
    "PYL": 'O'
}

global FULL_NAME
FULL_NAME = {
    "ALANINE": 'ALA',
    "CYSTEINE": 'CYS',
    "ASPARTIC ACID": 'ASP',
    "ASPARTATE": 'ASP',
    "GLUTAMIC ACID": 'GLU',
    "GLUTAMATE": 'GLU',
    "PHENYLALANINE": 'PHE',
    "GLYCINE": 'GLY',
    "HISTIDINE": 'HIS',
    "ISOLEUCINE": 'ILE',
    "LYSINE": 'LYS',
    "LEUCINE": 'LEU',
    "METHIONINE": 'MET',
    "ASPARAGINE": 'ASN',
    "PYRROLYSINE": 'PYL',
    "PROLINE": 'PRO',
    "GLUTAMINE": 'GLN',
    "ARGININE": 'ARG',
    "SERINE": 'SER',
    "THREONINE": 'THR',
    "SELENOCYSTEINE": 'SEC',
    "VALINE": 'VAL',
    "TRYPTOPHAN": 'TRP',
    "TYROSINE": 'TYR',
}



import pkg_resources
import json
import apalib.apalibExceptions

class Data:
    def __init__(self):
        _stream = pkg_resources.resource_stream(__name__, 'json/chemistry.json')
        self._jData = json.load(_stream)

    class MutationError(Exception):
        pass

    def __setattr__(self, key, value):
        if key == '_jdata' and self._jData is not None:
            raise self.MutationError(f"Attempt to mutate or reassign a protected variable: {key}")
        else:
            self.__dict__[key] = value

    def GetJson(self):
        return self._jData

    def Map(self, dtype, val):
        val = self.Standardize(val)
        return self._jData["Map"][dtype][val]

    def ValidateAA(self, res):
        check_res = self.Standardize(res)
        if len(res) != 3:
            return False
        if check_res in self._jData["Map"]["Amino Acids"].keys():
            return True
        return False

    def ValidateDNA(self, res, no_u = False):
        check_res = self.Standardize(res)
        if len(res) != 2:
            return False
        if check_res == "-" or check_res == ".":
            return False
        if no_u and (check_res == "U" or check_res == "DU"):
            return False
        if check_res in self._jData["Map"]["DNA Nucleotides"].keys():
            return True
        return False

    def SetDNAName(self, name):
        return "D" + self.Map("DNA Nucleotides", name)[0]

    def ValidateRNA(self, res, no_t = False):
        check_res = self.Standardize(res)
        if len(res) != 1:
            return False
        if check_res == "-" or res == ".":
            return False
        if no_t and check_res == "T":
            return False
        if check_res in self._jData["Map"]["DNA Nucleotides"].keys():
            return True
        return False

    def SetRNAName(self, name):
        return self.Map("RNA Nucleotides", name)[0]

    def Standardize(self, res):
        ret_str = ""
        for i in range(len(res)):
            if i == 0:
                ret_str += res[i].upper()
            elif i == 1 and len(res) == 2:
                ret_str += res[i].upper()
            else:
                ret_str += res[i].lower()
        return ret_str

    def _BuildConfiguration(self, element):
        ret = self._jData['Atoms'][element]["Ground Configuration"]
        while ret.find(']') != -1:
            next = ret[1:ret.find(']')]
            ret = ret[ret.find(']') + 1:]
            ret = self._jData['Atoms'][self.Map('Elements',next)]["Ground Configuration"] + " " + ret
        return ret


data = Data()



from setuptools import setup, find_packages
classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Science/Research',
    'Operating System :: Microsoft :: Windows :: Windows 10',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3'
]

setup(
    name='apalib',
    version='0.0.7',
    # description='Advanced Protein Analysis library',
    # long_description=open('README.md').read() +'\n\n' + open('CHANGELOG.txt').read(),
    author='Nate McMurray',
    author_email='natemresearch@gmail.com',
    license='MIT',
    classifiers=classifiers,
    keywords='Protein',
    packages=find_packages(),
    install_requires=[],
    include_package_data=True,
    package_data={'': ['data/*.json']},
    python_requires='>=3',
    url='https://github.com/cathepsin/apalib'
)


import apalib.config
import apalib.apalibExceptions
from apalib.Container import Container
from apalib.AminoAcid import AminoAcid
from apalib.Atom import Atom
from apalib.DNA import DNA
from apalib.RNA import RNA
from apalib.HETATM import HETATM
from apalib.pdb import PDB
from apalib.Data import Data
import math



# TODO work on error messages
# try:
#     requests.get(r"http://www.google.com", timeout=5)
# except :
#     sys.stderr.write(apalib.apalibExceptions.BadInternet())


# TODO Move **ALL HARDCODED DATA** into a json
# TODO Make error messages a full string (functional with stderr.write())
# TODO get this to work with DNA and RNA domains https://pdb101.rcsb.org/learn/guide-to-understanding-pdb-data/primary-sequences-and-the-pdb-format
# Parsing DNA/RNA should be essentially the same as an amino acid atom-group, with some intricacies. 1pyi is a good example
# TODO Check other PDB standards
# TODO Ensure functionality with symmetry pairs
# TODO Make set_name a thing everywhere

#  Naive approach to calculating isoelectric point at a specific pH

def getIEP(pHofInterest):
    global CONTAINER
    curr_protein_chains = CONTAINER.PeptideChains

    # initialize dict of counts
    total_counts = {'NTERM': 0, 'CTERM': 0, 'CYS': 0, 'ASP': 0, 'GLU': 0, 'HIS': 0, 'LYS': 0, 'ARG': 0, 'TYR': 0}
    import copy
    for key, value in curr_protein_chains.items():
        current_chain = value
        #TODO Add check for termini (in case section of a protein is checked instead of a full protein)
        #TODO Are there pKas for nucleic acids?

        total_counts['NTERM'] += 1
        total_counts['CTERM'] += 1

        for key, value in current_chain.items():
            if value.name == "CYS":
                total_counts['CYS'] += 1
            elif value.name == "ASP":
                total_counts['ASP'] += 1
            elif value.name == "GLU":
                total_counts['GLU'] += 1
            elif value.name == "HIS":
                total_counts['HIS'] += 1
            elif value.name == "LYS":
                total_counts['LYS'] += 1
            elif value.name == "ARG":
                total_counts['ARG'] += 1
            elif value.name == "TYR":
                total_counts['TYR'] += 1
    print(total_counts)

    # TODO Add pKa values to JSON file
    # using "EMBOSS" pKa's

    sum_charges = 0

    if pHofInterest < 8.6:
        sum_charges += total_counts['NTERM']
    if pHofInterest > 3.6:
        sum_charges -= total_counts['CTERM']
    if pHofInterest > 8.5:
        sum_charges -= total_counts['CYS']
    if pHofInterest > 3.9:
        sum_charges -= total_counts['ASP']
    if pHofInterest > 4.1:
        sum_charges -= total_counts['GLU']
    if pHofInterest < 6.5:
        sum_charges += total_counts['HIS']
    if pHofInterest < 10.8:
        sum_charges += total_counts['LYS']
    if pHofInterest < 12.5:
        sum_charges += total_counts['ARG']
    if pHofInterest > 10.1:
        sum_charges -= total_counts['TYR']

    return sum_charges










# TODO in all GETTERS in each class, add a check to see if variable exists
# TODO Complete the below functions before pushing version 1.0.0

# def SurfaceArea(**kwargs):
#     if 'domain' not in kwargs:
#         sys.stderr.write("No domain provided. Specify using SurfaceArea(domain=val)\n")
#     if 'num_dots' not in kwargs:
#         sys.stderr.write("Number of dots not provided. Specify using SurfaceArea(num_dots=val)\n")
#     if 'solvent_radius' not in kwargs:
#         sys.stderr.write("Solvent radius not provided. Specify using SurfaceArea(solvent_radius=val)\n")
#     print("stub")
#

def VectorPair(AA1, AA2):
    # Slope of line made from CA --> CENTROID
    CA1 = AA1.GetCA()
    CA2 = AA2.GetCA()
    # XY orthogonal projection
    m1 = (AA1.GetCentroid()[1] - CA1.GetCoordinates()[1]) / (AA1.GetCentroid()[0] - CA1.GetCoordinates()[0])
    m2 = (AA2.GetCentroid()[1] - CA2.GetCoordinates()[1]) / (AA2.GetCentroid()[0] - CA2.GetCoordinates()[0])
    '''
     Intersection point
                   y_2 - y_1 + m_1x_1 - m_2x_2
       x_int =    ------------------------------
                           m_1 - m_2
    '''
    x_int = (AA2.GetCentroid()[1] - AA1.GetCentroid()[1] + m1 * AA1.GetCentroid()[0] - m2 * AA2.GetCentroid()[0])
    ''' 
    Cast back into 3D
            x - x_1
       z = --------- * c + z_1
               a
    '''
    y1 = (x_int - AA1.GetCentroid()[0]) / AA1.vector[0] * AA1.vector[1] + AA1.GetCentroid()[1]
    y2 = (x_int - AA2.GetCentroid()[0]) / AA2.vector[0] * AA2.vector[1] + AA2.centroid[1]
    z1 = (x_int - AA1.GetCentroid()[0]) / AA1.vector[0] * AA1.vector[2] + AA1.GetCentroid()[2]
    z2 = (x_int - AA2.GetCentroid()[0]) / AA2.vector[0] * AA2.vector[2] + AA2.GetCentroid()[2]
    return abs(z1 - z2), [x_int, y1, z1], [x_int, y2, z2]

def GetCentAngle(res1, res2, rad = True):
    #Get angle using res1_centroid --> res1_CA --> res2_CA
    CA1 = res1.GetCA().GetCoordinates()
    CA2 = res2.GetCA().GetCoordinates()
    cent = res1.GetCentroid()
    a = GetDist(cent, CA1)
    b = GetDist(CA1, CA2)
    c = GetDist(cent, CA2)
    theta = math.acos(-1 * (c**2 - a**2 - b**2)/(2*a*b))
    if rad:
        return theta
    return math.degrees(theta)

#Checks that a residue-vector does not point backwards to extend towards a point
#TODO test this function a bit more
def CheckIsForward(res, pt):
    v1 = res.GetVector()
    v2 = [res.centroid[0] - pt[0],
          res.centroid[1] - pt[1],
          res.centroid[2] - pt[2]]
    #If any of the components switch direction then all v1_i * v2_i will be negative
    if v1[0] * v2[0] < 0 or v1[1] * v2[1] < 0 or v1[2] * v2[2] < 0:
        return False
    return True

def GetDist(a1, a2):
    return ((a1[0] - a2[0]) ** 2 + (a1[1] - a2[1]) ** 2 + (a1[2] - a2[2]) ** 2) ** (1 / 2)
#
# def CheckBridge(object1, object2):
#     print("stub")

# TODO Make sure ALL flags are properly cleared automatically




import sys
import apalib.config as config
import apalib.apalibExceptions as apaExcept
from apalib.Data import data as data
global FLAGS
FLAGS = {}


class AminoAcid:
    def __init__(self, seqNum = None, atoms = None, resName = None, rotamer = None, vector = None, set_name = True,
                 heptad = None, centroid = None, chainID=None):
        self.rotamer = rotamer
        self.vector = vector
        self.heptad = heptad
        self.centroid = centroid
        self.SetResName(resName, set_name)
        self.SetSeqNum(seqNum)
        self.SetAtoms(atoms)
        self.SetChainID(chainID)

    def DeepCopy(self):
        return AminoAcid(number=self.seqNum,
                         atoms=self.atoms,
                         name=self.resName,
                         rotamer=self.rotamer,
                         vector=self.vector,
                         heptad=self.heptad,
                         centroid=self.centroid,
                         chainID=self.chainID,
                         set_name=True)

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetSeqNum(self, num):
        self.seqNum = num

    def SetAtoms(self, atoms):
        self.atoms = atoms
        # self.CalculateCentroid()

    def InsertAtom(self, atom):
        if self.atoms is None:
            self.atoms = list()
        self.atoms.append(atom)

    def GetAtoms(self):
        return self.atoms

    def GetRotamers(self, **kwargs):
        accepted = ['unique']
        for key in kwargs.keys():
            if key not in accepted:
                raise apalib.apalibExceptions.BadKwarg(accepted)

        if self.rotamer is not None:
            return self.rotamer

        if 'unique' in kwargs.keys() and isinstance(kwargs['unique'], (bool)):
            unique = kwargs['unique']
        else:
            unique = False

        retDict = {'Common':[]}
        if self.atoms is not None:
            for atom in self.atoms:
                if atom.rotation is None or atom.rotation == '':
                    retDict['Common'].append(atom)
                    continue
                if atom.rotation not in retDict.keys():
                    retDict[atom.rotation] = []
                retDict[atom.rotation].append(atom)
        if unique is False:
            for lst in [retDict[key] for key in retDict.keys() if key != 'Common']:
                lst += retDict['Common']
        return retDict

    def GetCA(self):
        if self.atoms is None:
            return None
        for atom in self.atoms:
            if atom.GetName() == 'CA':
                return atom
        return None

    def SetResName(self, name, set_name):
        if name is None:
            self.resName = name
            return
        if not set_name:
            self.resName = name
            self.RaiseFlag('NO_NAME_CHECK')
            return
        else:
            self.ClearFlag('NO_NAME_CHECK')
        if len(name) == 3 and name in THREE_LETTER:
            self.resName = name
        elif len(name) == 3 and name not in THREE_LETTER:
            self.resName = None
        elif len(name) == 1:
            self.resName = self.OneToThree(name)
        elif len(name) <= 4:
            if name[-3:] in THREE_LETTER:
                self.resName = name[-3:]
                self.rotamer = name[:-3]
            else:
                self.resName = None
        elif len(name) == 2:
            self.resName = None
        if self.resName is not None:
            self.RaiseFlag('BAD_NAME')
        else:
            self.RaiseFlag('BAD_NAME')
        if self.resName is not None and self.rotamer is not None:
            self.RaiseFlag('MARKED')
        elif self.resName is not None and self.rotamer is None:
            self.ClearFlag('MARKED')

    def SetHeptad(self, heptad):
        self.heptad = heptad

    def CalculateCentroid(self):
        if 'atoms' not in self.__dict__:
            self.centroid = None
            return
        #TODO switch to using the JSON for this
        # *For Glycine, only the alpha carbon is considered
        # *For Alanine, only the beta carbon is considered
        # AAs = {
        #     "SER": ['OG'],
        #     "CYS": ['SG'],
        #     "SEC": ['SE'],
        #     "GLY": ['CA'],
        #     "ALA": ['CB'],
        #     "THR": ['OG1', 'CG2'],
        #     "PRO": ['CG', 'CD'],
        #     "VAL": ['CG1', 'CG2'],
        #     "ASP": ['CG', 'OD1', 'OD2'],
        #     "ASN": ['CG', 'OD1', 'ND2'],
        #     "ILE": ['CG1', 'CG2', 'CD1'],
        #     "LEU": ['CG', 'CD1', 'CD2'],
        #     "MET": ['CG', 'SD', 'CE'],
        #     "LYS": ['CG', 'CD', 'CE', 'NZ'],
        #     "GLU": ['CG', 'CD', 'OE1', 'OE2'],
        #     "GLN": ['CG', 'CD', 'OE1', 'NE2'],
        #     "HIS": ['CG', 'ND1', 'CD2', 'CE1', 'NE2'],
        #     "ARG": ['CG', 'CD', 'NE', 'CZ', 'NH1', 'NH2'],
        #     "PHE": ['CG', 'CD1', 'CD2', 'CE1', 'CE2', 'CZ'],
        #     "TYR": ['CG', 'CD1', 'CD2', 'CE1', 'CE2', 'CZ', 'OH'],
        #     "TRP": ['CG', 'CD1', 'CD2', 'NE1', 'CE2', 'CE3', 'CZ2', 'CZ3', 'CH2']
        # }
        # TODO Deal with non-1 occupancy
        residue = data.Map('Amino Acids', self.resName)
        if residue in data.GetJson()['Amino Acids'].keys():
            # If residue is not specified
            x_coord = 0
            y_coord = 0
            z_coord = 0
            num_atoms = 0
            for atom in [atom for atom in self.atoms if atom.name in data.GetJson()['Amino Acids'][residue]['Centroid']]:
                if atom.element == 'H':
                    continue
                x_coord += atom.GetCoordinates()[0]
                y_coord += atom.GetCoordinates()[1]
                z_coord += atom.GetCoordinates()[2]
                num_atoms += 1
            self.centroid = list()
            try:
                self.centroid.append(x_coord / num_atoms)
                self.centroid.append(y_coord / num_atoms)
                self.centroid.append(z_coord / num_atoms)
            except ZeroDivisionError:
                self.centroid.clear()
                beta = [atom for atom in self.atoms if atom.name == 'CB']
                alpha = [atom for atom in self.atoms if atom.name == 'CA']
                if len(beta) != 0:
                    self.centroid.append(beta[0].coordinates[0])
                    self.centroid.append(beta[0].coordinates[1])
                    self.centroid.append(beta[0].coordinates[2])
                    self.RaiseFlag('B_CENTROID')
                elif len(alpha) != 0:
                    self.centroid.append(alpha[0].coordinates[0])
                    self.centroid.append(alpha[0].coordinates[1])
                    self.centroid.append(alpha[0].coordinates[2])
                    self.RaiseFlag('A_CENTROID')
                else:
                    self.centroid = None
                    self.vector = None
                    self.RaiseFlag('BAD_CENTROID')
                    return
            # After all that, set the centroidal vector
            self.vector = [self.centroid[0] - self.GetCA().GetCoordinates()[0],
                           self.centroid[1] - self.GetCA().GetCoordinates()[1],
                           self.centroid[2] - self.GetCA().GetCoordinates()[2]]

        elif len(self.resName) == 4 and self.resName[1:] in data.GetJson()['Amino Acids'][residue]['Centroid']:
            sys.stderr.write("ROTAMER!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        else:
            sys.exit("UNKNOWN AMINO ACID YO. DO SOMETHING ABOUT THIS!!!!!")

    def GetCentroid(self):
        return self.centroid

    def GetVector(self):
        return self.vector

    def SetChainID(self, c):
        self.chainID = c

    def GetASA(self, form):
        d = config.data
        name = config.data.Map("Amino Acids", config.data.Standardize(self.resName))
        return config.data.GetJson()["Amino Acids"][name]["ASA"][form]

    def WriteForPDB(self):
        retstr = ""
        for atom in self.atoms:
            retstr += atom.WritePDB(intro="ATOM  ")
        return retstr

    @staticmethod
    def OneToThree(oln):
        if oln in ONE_LETTER:
            try:
                return ONE_LETTER[oln]
            except KeyError:
                return None

    @staticmethod
    def NameToThree(name):
        if name in FULL_NAME:
            try:
                return FULL_NAME[name]
            except KeyError:
                return None

    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False

    # TODO These would be pretty cool to implement if possible. Open and write to a new python file that these access?
    @staticmethod
    def Set_lt(str):
        print('stub')

    @staticmethod
    def Set_repr(str):
        print('stub')


    def Set_str(self, str):
        return

    def __lt__(self, other):
        return self.seqNum < other.number

    def __repr__(self):
        return f"RESIDUE: {self.resName}, NUMBER: {self.seqNum}"

    def __str__(self):
        return f"{self.resName} {self.seqNum}"


# Shoved down here for cleanliness
global ONE_LETTER
ONE_LETTER = {
    "R": 'ARG',
    "H": 'HIS',
    "K": 'LYS',
    "D": 'ASP',
    "E": 'GLU',
    "S": 'SER',
    "T": 'THR',
    "N": 'ASN',
    "Q": 'GLN',
    "C": 'CYS',
    "U": 'SEC',
    "G": 'GLY',
    "P": 'PRO',
    "A": 'ALA',
    "V": 'VAL',
    "I": 'ILE',
    "L": 'LEU',
    "M": 'MET',
    "F": 'PHE',
    "Y": 'TYR',
    "W": 'TRP',
    "O": 'PYL'
}

global THREE_LETTER
THREE_LETTER = {
    "ARG": 'R',
    "HIS": 'H',
    "LYS": 'K',
    "ASP": 'D',
    "GLU": 'E',
    "SER": 'S',
    "THR": 'T',
    "ASN": 'N',
    "GLN": 'G',
    "CYS": 'C',
    "SEC": 'U',
    "GLY": 'G',
    "PRO": 'P',
    "ALA": 'A',
    "VAL": 'V',
    "ILE": 'I',
    "LEU": 'L',
    "MET": 'M',
    "PHE": 'F',
    "TYR": 'Y',
    "TRP": 'W',
    "PYL": 'O'
}

global FULL_NAME
FULL_NAME = {
    "ALANINE": 'ALA',
    "CYSTEINE": 'CYS',
    "ASPARTIC ACID": 'ASP',
    "ASPARTATE": 'ASP',
    "GLUTAMIC ACID": 'GLU',
    "GLUTAMATE": 'GLU',
    "PHENYLALANINE": 'PHE',
    "GLYCINE": 'GLY',
    "HISTIDINE": 'HIS',
    "ISOLEUCINE": 'ILE',
    "LYSINE": 'LYS',
    "LEUCINE": 'LEU',
    "METHIONINE": 'MET',
    "ASPARAGINE": 'ASN',
    "PYRROLYSINE": 'PYL',
    "PROLINE": 'PRO',
    "GLUTAMINE": 'GLN',
    "ARGININE": 'ARG',
    "SERINE": 'SER',
    "THREONINE": 'THR',
    "SELENOCYSTEINE": 'SEC',
    "VALINE": 'VAL',
    "TRYPTOPHAN": 'TRP',
    "TYROSINE": 'TYR',
}



# global FULL_NAME
# global ONE_LETTER
global FLAGS
FLAGS = {}

class RNA:
    def __init__(self, **kwargs):
        self.number = None
        self.atoms = None
        self.name = None

        set_name = True
        if 'set_name' in kwargs:
            set_name = kwargs['set_name']
        if 'name' in kwargs:
            self.SetName(kwargs['name'], set_name)
        if 'number' in kwargs:
            self.SetNumber(kwargs['number'])
        if 'atoms' in kwargs:
            self.SetAtoms(kwargs['atoms'])

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetNumber(self, num):
        self.number = num

    def SetAtoms(self, atoms):
        self.atoms = atoms
        self.GetCentroid(atoms)

    def InsertAtom(self, atom):
        if self.atoms is None:
            self.atoms = list()
        self.atoms.append(atom)

    def SetName(self, name, set_name):
        if not set_name:
            self.name = name
            self.RaiseFlag('NO_NAME_CHECK')
            self.ClearFlag('BAD_NAME')
            return
        else:
            self.ClearFlag('NO_NAME_CHECK')
        global ONE_LETTER
        if name in ONE_LETTER:
            self.name = ONE_LETTER[name]
        elif name in FULL_NAME:
            self.name = FULL_NAME[name]
        else:
            self.RaiseFlag('BAD_NAME')
            return
        self.ClearFlag('BAD_NAME')
        return

    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False

    def __lt__(self, other):
        return self.number < other.number

    def __repr__(self):
        return f"RESIDUE: {self.name}, NUMBER: {self.number}"

    def __str__(self):
        return f"{self.name} {self.number}"

global ONE_LETTER
ONE_LETTER = {
    'A': 'ADENINE',
    'C': 'CYTOSINE',
    'G': 'GUANINE',
    'T': 'THYMINE',
    'U': 'URACIL',
    'I': 'INOSINE'
}
global FULL_NAME
FULL_NAME = {
    'ADENINE': 'A',
    'CYTOSINE': 'C',
    'GUANINE': 'G',
    'THYMINE': 'T',
    'URACIL': 'U',
    'INOSINE': 'I'
}



from unittest import TestCase


class TestDNA(TestCase):
    def test_set_number(self):
        self.fail()

    def test_set_atoms(self):
        self.fail()

    def test_insert_atom(self):
        self.fail()

    def test_set_name(self):
        self.fail()

    def test_set_n(self):
        self.fail()



import pkg_resources
import json

class Data:
    def __init__(self):
        _stream = pkg_resources.resource_stream(__name__, 'data/chemistry.json')
        self._jData = json.load(_stream)

    def GetJson(self):
        return self._jData

    def Map(self, dtype, val):
        return self._jData["Map"][dtype][val]

    def ValidateAA(self, res):
        check_res = self.Standardize(res)
        if check_res in self._jData["Map"]["Amino Acids"].keys():
            return True
        return False

    def ValidateDNA(self, res, no_u = False):
        check_res = self.Standardize(res)
        if check_res == "-" or check_res == ".":
            return False
        if no_u and (check_res == "U" or check_res == "DU"):
            return False
        if check_res in self._jData["Map"]["DNA Nucleotides"].keys():
            return True
        return False

    def ValidateRNA(self, res, no_t = False):
        check_res = self.Standardize(res)
        if check_res == "-" or res == ".":
            return False
        if no_t and check_res == "T":
            return False
        if check_res in self._jData["Map"]["DNA Nucleotides"].keys():
            return True
        return False

    def Standardize(self, res):
        ret_str = ""
        for i in range(len(res)):
            if i == 0:
                ret_str += res[i].upper()
            elif i == 1 and len(res) == 2:
                ret_str += res[i].upper()
            else:
                ret_str += res[i].lower()
        return ret_str



import apalib.apalibExceptions
import apalib.apalibExceptions as apaExcept
import sys

from apalib import *

class PDB:
    def __init__(self):
        self.container = Container()

    def Contents(self):
        return self.container

    def Fetch(self, prot):
        # print("Fetching ", prot)
        import urllib.request
        url = r'https://files.rcsb.org/download/' + prot.strip() + '.pdb'
        try:
            with urllib.request.urlopen(url) as f:
                self.container.SetFetch(f.read().decode('utf-8'))
                self._Parse()
        except urllib.error.URLError:
            sys.stderr.write("The requested pdb code could not be retrieved or does not exist\n")

    def Read(self, path):
        with open(path, 'r') as fp:
            self.container.SetFetch(fp.read())
            self._Parse()

    # Wrapper for the ParsePDB file to allow functionality with a fetched protein
    def _Parse(self):
        try:
            if self.container.GetFetch() is None:
                raise apaExcept.NoFetchError
            return self._ParsePDB(self.container.GetFetch().splitlines())
        except apaExcept.NoFetchError as e:
            sys.stderr.write(e.message)

    def _ParsePDB(self, pdbFile):
        proteinChains = dict()
        DNAChains = dict()
        RNAChains = dict()
        HETATMChains = dict()
        for line in pdbFile:
            if line[0:4] == "ATOM" or line[0:6] == 'HETATM':
                # Create a new chain if needed
                chain = line[20:22].strip()
                if chain == '':
                    chain = '$'
                groupNumber = int("".join(chr for chr in line[22:27] if chr.isdigit()))
                atomNumber = int(line[6:11])
                id = line[11:16].strip()
                groupName = line[16:20].strip()
                coordinates = [line[27:38], line[38:46], line[46:55]]
                occupancy = float(line[55:60])
                bfactor = float(line[60:67])
                element = line[77:].strip()
                newAtom = config.Atom(number=atomNumber, coordinates=coordinates, id=id, occupancy=occupancy,
                               b_factor=bfactor, element=element, residue=groupName)

                # RNA atom
                if line[0:6] == "HETATM":
                    if chain not in HETATMChains:
                        HETATMChains[chain] = dict()
                    if groupNumber not in HETATMChains[chain]:
                        HETATMChains[chain][groupNumber] = config.HETATM(name=groupName, number=groupNumber)
                    HETATMChains[chain][groupNumber].InsertAtom(newAtom)
                elif len(groupName) == 1:
                    if chain not in RNAChains:
                        RNAChains[chain] = dict()
                    if groupNumber not in RNAChains[chain]:
                        RNAChains[chain][groupNumber] = config.RNA(name=groupName, number=groupNumber)
                    RNAChains[chain][groupNumber].InsertAtom(newAtom)

                elif len(groupName) == 2:
                    if chain not in DNAChains:
                        DNAChains[chain] = dict()
                    if groupNumber not in DNAChains[chain]:
                        DNAChains[chain][groupNumber] = config.DNA(name=groupName, number=groupNumber)
                    DNAChains[chain][groupNumber].InsertAtom(newAtom)

                elif len(groupName) == 3 or len(groupName) == 4:
                    if chain not in proteinChains:
                        proteinChains[chain] = dict()
                    if groupNumber not in proteinChains[chain]:
                        proteinChains[chain][groupNumber] = config.AminoAcid(name=groupName, number=groupNumber)
                    proteinChains[chain][groupNumber].InsertAtom(newAtom)

        # Postparsing functions
        self.container.SetProteinChains(proteinChains)
        self.container.SetDNAChains(DNAChains)
        self.container.SetRNAChains(RNAChains)
        self.container.SetHETATMChains(HETATMChains)

        for chain in self.container.GetPeptideChains():
            for index in self.container.GetPeptideChains()[chain]:
                self.container.GetPeptideChains()[chain][index].CalculateCentroid(self.container.GetPeptideChains()[chain][index].atoms)

    # def CountResidues(self, **kwargs):
    #     for key in kwargs:
    #         if key != 'find' or (key == 'find' and not isinstance(kwargs['find'], list)):
    #             raise apalib.apalibExceptions.BadKwarg('find=[residue_name1, residue_name2, ...]')
    #
    #     # If a specific residue is wanted
    #     if 'find' in kwargs:
    #
    #     else:
    #
    # #TODO GetMass() of a current fetch
    #
    #
    # def __StandardizeResidue(self, res, **kwargs):
    #     accepted_kwargs = ['']
    #


    def Validate(self, **kwargs):
        for key in kwargs:
            if key != 'pdb' or (key == 'pdb' and not isinstance(kwargs['pdb'], str)):
                raise apalib.apalibExceptions.BadKwarg('pdb=<pdb_to_validate>')


from unittest import TestCase


class TestAminoAcid(TestCase):
    def test_set_number(self):
        self.fail()

    def test_set_atoms(self):
        self.fail()

    def test_insert_atom(self):
        self.fail()

    def test_insert_atom(self):
        self.fail()

    def test_get_atoms(self):
        self.fail()

    def test_set_name(self):
        self.fail()

    def test_set_heptad(self):
        self.fail()

    def test_calculate_centroid(self):
        self.fail()

    def test_get_centroid(self):
        self.fail()

    def test_one_to_three(self):
        self.fail()

    def test_name_to_three(self):
        self.fail()



from apalib.Container import Container
import apalib.Container
from apalib.AminoAcid import AminoAcid
from apalib.Atom import Atom
from apalib.DNA import DNA
from apalib.RNA import RNA
from apalib.HETATM import HETATM

from apalib.Data import Data
Container = Container()
data = Data()


from unittest import TestCase
import __init__

class Testinit(TestCase):
    def setUp(self):
        __init__.Fetch('5u59')
        print("Setup complete")

    def test_set_number(self):
        self.fail()

    def test_set_atoms(self):
        self.fail()

    def test_insert_atom(self):
        self.fail()

    def test_set_name(self):
        self.fail()



import apalib

pdb = apalib.PDB()
pdb.Fetch('7nyr')
print("Stop")


























import apalib

# pdb = apalib.PDB()
# pdb.FetchAsFile('5u59')
# print(pdb.FetchFASTA('5u59'))
# print("Stop")
#
#
#
#

import apalib

pdb = apalib.PDB()
pdb.Fetch('5u59')
print("stop")











# data = apalib.Data()
#
# # with open(r"apalib/data/chemistry.json", "w") as f:
# #     json.dump(data.GetJson(), f, indent=3)
#
#
# pdb = apalib.PDB()
# # pdb.Fetch('5u59') #Solid protein
# # pdb.Fetch('4kpy') #DNA-containing protein
# # pdb.Fetch('3wbm') #RNA-containing protein
# # pdb.WritePDB('testpdb.pdb')
# # pdb.RemoveWater()
# # pdb.WritePDB()
# # pdb.Contents().DumpAtoms()
# pdb.Read(r"C:\Users\natem\OneDrive\Desktop\Research\ccSASA - Copy\1j1j\1j1j.mmol")
# (apalib.GetUnfoldedSASA(pdb.Contents().GetPeptideChains()['A']))
#
# pdb.RemoveWater()
#
# print(pdb.Contents().GetPeptideChains()['A'][1].GetAtoms()[0].GetRadius())
# (pdb.Contents().GetPeptideChains()['A'][1].GetAtoms()[0].GetConfiguration(extended = True))
# # print(pdb.Contents().DumpAtoms())
#
# one = pdb.Contents().GetPeptideChains()['A'][1]
# two = pdb.Contents().GetPeptideChains()['A'][2]
#
# val = apalib.VectorPair(one, two)
# (val)
#
# (apalib.GetCentAngle(one, two, rad=False))
#
# data = apalib.Data()
# pdb.Read(r"C:\Users\natem\OneDrive\Desktop\Research\ccSASA - Copy\1j1j\1j1j.mmol")
# ("Stop")



# global ONE_LETTER
# global TWO_LETTER
# global FULL_NAME
from apalib.Data import data as data
global FLAGS
FLAGS = {}


class DNA:
    def __init__(self, seqNum=None, atoms=None, resName=None, set_name=True, chainID=None):
        # self.seqNum = number
        # self.atoms = atoms
        # self.resName = name
        self.SetResName(resName, set_name)
        self.SetSeqNum(seqNum)
        self.SetAtoms(atoms)
        self.SetChainID(chainID)

    def DeepCopy(self):
        return DNA(seqNum=self.seqNum,
                   atoms=self.atoms,
                   resName=self.resName,
                   set_name=True)

    def AddAttribute(self, attr, var):
        self.__dict__[attr] = var

    def SetSeqNum(self, num):
        self.seqNum = num

    def SetAtoms(self, atoms):
        self.atoms = atoms
        self.CalculateCentroid(atoms)

    def GetAtoms(self):
        return self.atoms

    def CalculateCentroid(self, atoms):
        print("DNA PRESENT! FILL THIS STUB")

    def InsertAtom(self, atom):
        if self.atoms is None:
            self.atoms = list()
        self.atoms.append(atom)

    def SetResName(self, name, set_name):
        if not set_name:
            self.resName = name
            self.RaiseFlag('NO_NAME_CHECK')
            return
        else:
            self.CheckFlag('NO_NAME_CHECK')
        if data.ValidateDNA(name):
            self.resName = data.SetDNAName(name)
        # elif name in self.FULL_NAME:
        #     self.resName = self.FULL_NAME[name]
        # elif name in self.ONE_LETTER:
        #     self.resName = 'D' + name
        else:
            self.resName = None
            self.RaiseFlag('BAD_NAME')
            return
        self.CheckFlag('BAD_NAME')
        return

    def GetResName(self):
        return self.resName

    def SetChainID(self, c):
        self.chainID = c

    def WriteForPDB(self):
        retstr = ""
        for atom in self.atoms:
            retstr += atom.WritePDB(intro="ATOM  ")
        return retstr


    @staticmethod
    def CheckFlag(f):
        global FLAGS
        if f in FLAGS:
            return FLAGS[f]
        return False

    @staticmethod
    def RaiseFlag(flag):
        global FLAGS
        FLAGS[flag] = True

    @staticmethod
    def ClearFlag(flag):
        global FLAGS
        FLAGS[flag] = False

    def __lt__(self, other):
        return self.seqNum < other.number

    def __repr__(self):
        return f"RESIDUE: {self.resName}, NUMBER: {self.seqNum}"

    def __str__(self):
        return f"{self.resName} {self.seqNum}"


global ONE_LETTER
ONE_LETTER = {
    'A': 'ADENINE',
    'C': 'CYTOSINE',
    'G': 'GUANINE',
    'T': 'THYMINE',
    'U': 'URACIL',
    'I': 'INOSINE'
}
global TWO_LETTER
TWO_LETTER = {
    'DA': 'ADENINE',
    'DC': 'CYTOSINE',
    'DG': 'GUANINE',
    'DT': 'THYMINE',
    'DU': 'URACIL',
    'DI': 'INOSINE'
}
global FULL_NAME
FULL_NAME = {
    'ADENINE': 'DA',
    'CYTOSINE': 'DC',
    'GUANINE': 'DG',
    'THYMINE': 'DT',
    'URACIL': 'DU',
    'INOSINE': 'DI'
}



# class ExceptionWrapper(Exception):
#     def __init__(self, message):
#         self.message = message
#
#     def __str__(self):
#         return self.message


class MutationError(Exception):
    # When a constant variable is mutated
    def __init__(self):
        self.message = 'Attempt to mutate a protected variable'


class NoFetchError(Exception):
    # Used when a fetched pdb file is required but not provided
    def __init__(self):
        self.message = 'No fetched protein. Use method Fetch(PDB_CODE) to fetch a protein.\n'
        return self.message

    def __str__(self):
        return self.message


class BadInternet(Exception):
    def __init__(self):
        self.message = 'Poor or no internet connection. Some functionality may be lost\n\t' \
                       r'Connection to http://www.google.com timeout or failure'

    def __str__(self):
        return self.message


class BadKwarg(Exception):
    def __init__(self, accepted):
        self.message = f'Bad parameter provided. Accepted parameters are as follows: {accepted}'

    def __str__(self):
        return self.message


class BadMethodException(Exception):
    pass

class NegativeProbabilityException(Exception):
    pass

class ProbabilityOverflowException(Exception):
    pass

class PolarProbabilityOverflowException(Exception):
    pass

class ChargedProbabilityOverflowException(Exception):
    pass

class ChargeLengthMismatchException(Exception):
    pass

class BadBatchSizeException(Exception):
    pass

class InvalidURLException(Exception):
    pass

class BadJSONException(Exception):
    pass

class InvalidPresetDistributionException(Exception):
    pass

class BadMaxAttemptsException(Exception):
    pass

class ImpossibleSettingsException(Exception):
    pass
